
# ------------------------------
# --- 3d point

class PPoint
  def initialize(x=0, y=0, z=0)
    @x = x
    @y = y
    @z = z
  end

  attr_accessor :x
  attr_accessor :y
  attr_accessor :z

  def add( v2 )
    @x += v2.x
    @y += v2.y
    @z += v2.z
  end

  def zoom( val )
    @x *= val
    @y *= val
    @z *= val
  end

  def length
    length = Math.sqrt( @x*@x + @y*@y + @z*@z )
  end

  def normalize
    l = length()
    return if l == 0
    p = 1.0 / l
    zoom( p )
  end

  def inner_product(v2)
    tv1 = PPoint.new(@x, @y, @z)
    tv2 = PPoint.new(v2.x, v2.y, v2.z)
    tv1.normalize
    tv2.normalize
    inner = (tv1.x * tv2.x) + (tv1.y * tv2.y) + (tv1.z * tv2.z)
  end

  def outer_product(v2)
    tv1 = PPoint.new(@x, @y, @z)
    tv2 = PPoint.new(v2.x, v2.y, v2.z)
    tv1.normalize
    tv2.normalize
    outer = PPoint.new((tv1.y * tv2.z) - (tv1.z * tv2.y),
                       (tv1.z * tv2.x) - (tv1.x * tv2.z),
                       (tv1.x * tv2.y) - (tv1.y * tv2.x))
  end

  def deg2rad( deg )
    rad = Math::PI * 2 * deg / 360.0
  end

  def rotate_x( deg )
    rx = deg2rad( deg )
    x = @x
    y = @y*Math.cos(rx) + @z*Math.sin(rx)
    z = -@y*Math.sin(rx) + @z*Math.cos(rx)
    @x = x
    @y = y
    @z = z
  end

  def rotate_y( deg )
    ry = deg2rad( deg )
    x = @x*Math.cos(ry) - @z*Math.sin(ry)
    y = @y
    z = @x*Math.sin(ry) + @z*Math.cos(ry)
    @x = x
    @y = y
    @z = z
  end

  def rotate_z( deg )
    rz = deg2rad( deg )
    x = @x*Math.cos(rz) + @y*Math.sin(rz)
    y = -@x*Math.sin(rz) + @y*Math.cos(rz)
    z = @z
    @x = x
    @y = y
    @z = z
  end

  def rotate( rot_x, rot_y, rot_z )
    rotate_x( rot_x )
    rotate_y( rot_y )
    rotate_z( rot_z )
  end

  def equal?( p2 )
    result = false
    length = PPoint.new(@x-p2.x, @y-p2.y, @z-p2.z).length.abs
    if ((p2.x == @x) && (p2.y == @y) && (p2.z == @z)) || 
       length < 0.00000001 then
      result = true
    end
    return result
  end
end

# ------------------------------
# --- 3d line

class PLine
  def initialize
    @point = []
  end

  def set_point( index, point )
    return if index > 2
    return if index < 0
    @point[index] = point
  end

  def get_point( index )
    @point[index]
  end

  def ok?
    @point.size == 2
  end

  def equal?(line)
    if @point[0].equal?(line.get_point(0)) &&
       @point[1].equal?(line.get_point(1)) then
      return true
    end
    if @point[0].equal?(line.get_point(1)) &&
       @point[1].equal?(line.get_point(0)) then
      return true
    end
    return false
  end

end


# ------------------------------
# --- 3d triangle

class PTriangle < PLine

  def set_point( index, point )
    return if index > 3
    return if index < 0
    @point[index] = point
  end

  def ok?
    @point.size == 3
  end

  def normalize
    v1 = PPoint.new(@point[1].x - @point[0].x,
                    @point[1].y - @point[0].y,
                    @point[1].z - @point[0].z)
    v2 = PPoint.new(@point[2].x - @point[0].x,
                    @point[2].y - @point[0].y,
                    @point[2].z - @point[0].z)
    outer = v1.outer_product(v2)
    if outer.z < 0 then
      p1 = @point[1]
      p2 = @point[2]
      @point.delete(p1)
      @point.delete(p2)
      @point << p2
      @point << p1
    end
  end

  def include?( point )
    if @point[0].equal?(point) then
      return true
    end
    if @point[1].equal?(point) then
      return true
    end
    if @point[2].equal?(point) then
      return true
    end
    #
    v1 = PPoint.new(@point[1].x - @point[0].x,
                    @point[1].y - @point[0].y,
                    @point[1].z - @point[0].z)
    v2 = PPoint.new(point.x - @point[0].x,
                    point.y - @point[0].y,
                    point.z - @point[0].z)
    outer = v1.outer_product(v2)
    z1 = outer.z
    v1 = PPoint.new(@point[2].x - @point[1].x,
                    @point[2].y - @point[1].y,
                    @point[2].z - @point[1].z)
    v2 = PPoint.new(point.x - @point[1].x,
                    point.y - @point[1].y,
                    point.z - @point[1].z)
    outer = v1.outer_product(v2)
    z2 = outer.z
    v1 = PPoint.new(@point[0].x - @point[2].x,
                    @point[0].y - @point[2].y,
                    @point[0].z - @point[2].z)
    v2 = PPoint.new(point.x - @point[2].x,
                    point.y - @point[2].y,
                    point.z - @point[2].z)
    outer = v1.outer_product(v2)
    z3 = outer.z
    return (z1 >= 0) && (z2 >= 0) && (z3 >= 0)
  end

end
